/*
*
*
*  Simple Interpreter (lib.c)  
*  Mafice <mafice@mflabo.net>
*
*
*/




#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <Resea/types.h>
#include <Resea/io.h>
#include <Resea/interrupt.h>


/*  IO ポート  */
#define IOPORT_KBC1             0x60
#define IOPORT_KBC2             0x64

/*  キーコード  */
#define KEY_ESC         0x01
#define KEY_BACKSPACE   0x0e
#define KEY_TAB         0x0f
#define KEY_ENTER       0x1c
#define KEY_HANKAKU     0x29
#define KEY_HIRAGANA    0x70
#define KEY_SPACE       0x39
#define KEY_SYSREQ      0x54
#define KEY_ALT         0x38
#define KEY_CTRL        0x1d
#define KEY_R_SHIFT     0x36
#define KEY_L_SHIFT     0x2a
#define KEY_CAPSLOCK    0x3a
#define KEY_NUMLOCK     0x45
#define KEY_SCROLLLOCK  0x46
#define KEY_F1          0x3b
#define KEY_F2          0x3c
#define KEY_F3          0x3d
#define KEY_F4          0x3e
#define KEY_F5          0x3f
#define KEY_F6          0x40
#define KEY_F7          0x41
#define KEY_F8          0x42
#define KEY_F9          0x43
#define KEY_F10         0x44
#define KEY_F11         0x57
#define KEY_F12         0x58



static unsigned char kbmap1[] = {
    0, KEY_ESC, '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '-', '^', KEY_BACKSPACE,
    KEY_TAB, 'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P', '@', '[', KEY_ENTER, KEY_CTRL,
    'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', ';', ':', KEY_HANKAKU, KEY_L_SHIFT, ']',
    'Z', 'X', 'C', 'V', 'B', 'N', 'M', ',', '.', '/', KEY_R_SHIFT, '*',
    KEY_ALT, KEY_SPACE, KEY_CAPSLOCK, KEY_F1, KEY_F2, KEY_F3, KEY_F4,
    KEY_F5, KEY_F6, KEY_F7, KEY_F8, KEY_F9, KEY_F10, KEY_NUMLOCK, KEY_SCROLLLOCK,
    '7', '8', '9', '-', '4', '5', '6', '+', '1', '2', '3', '0', '.', KEY_SYSREQ, 0, 0, KEY_F11, KEY_F12,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    KEY_HIRAGANA, 0, 0, '_', 0, 0, 0, 0, 0, 0, 0, 0, 0, '\\', 0, 0
};

static unsigned char kbmap2[] = {
    0, 0, '!', '"', '#', '$', '%', '&', 0, '(', ')', 0, '=', '~', 0, 0,
    'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P', '`', '{', 0, 0,
    'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', '+', '*', 0, 0, '}',
    'Z', 'X', 'C', 'V', 'B', 'N', 'M', '<', '>', '?', 0, '*',
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    '7', '8', '9', '-', '4', '5', '6', '+', '1', '2', '3', '0', '.', 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, '|', 0, 0, 0, 0, 0, 0, 0, 0, 0, '|', 0, 0

};

struct KeyboardInfo{
  char  CapsLock;
  char  NumLock;
  char  ScrollLock;
  char  alt;
  char  ctrl;
  char  shift;
};


char keycode2ascii (unsigned char keycode);
void keyboardLED (unsigned char target);
void waitKBC (int bit);

struct KeyboardInfo Keyboard;

char keybuf=0;

void keyboardISR (void){

 u8   keycode;
 char asccode;

  out8(0x20, 0x61);
  keycode  = in8(0x60);
  asccode  = keycode2ascii(keycode);

  if(asccode == 0)
      return;

  keybuf = asccode;

  return;
}




/*
*  Function: keycode2ascii
*
*    キーコードをASCIIコードに変換する
*
*  Parameters:
*
*    keycode - キーコード
*
*  Returns:
*
*    変換可能な文字の場合はそのASCIIコード、不可能な文字の場合は0を返します。
*
*/

char keycode2ascii (unsigned char keycode){

 char c;


  if(keycode == KEY_CTRL+0x80)
    Keyboard.ctrl      = 0;
  if(keycode == KEY_ALT+0x80)
    Keyboard.alt       = 0;
  if(keycode == KEY_L_SHIFT+0x80 || keycode == KEY_R_SHIFT+0x80)
    Keyboard.shift     = 0;

  if(keycode == KEY_CTRL){
    Keyboard.ctrl  = 1; return 0;
  }
  if(keycode == KEY_ALT){
    Keyboard.alt   = 1; return 0;
  }
  if(keycode == KEY_L_SHIFT || keycode == KEY_R_SHIFT){
    Keyboard.shift = 1; return 0;
  }

  if(keycode == KEY_ENTER)
    return '\n';

  if(keycode == KEY_SPACE)
    return ' ';

  if(keycode == KEY_TAB)
    return '\t';

  if(keycode == KEY_CAPSLOCK){
    keyboardLED(KEY_CAPSLOCK);

    if(Keyboard.CapsLock == 1)
      Keyboard.CapsLock = 0;
    else
      Keyboard.CapsLock = 1;

    return 0;
  }


  if(keycode == KEY_NUMLOCK){
    keyboardLED(KEY_NUMLOCK);

    if(Keyboard.NumLock == 1)
      Keyboard.NumLock = 0;
    else
      Keyboard.NumLock = 1;

    return 0;
  }

  if(keycode == KEY_SCROLLLOCK){
    keyboardLED(KEY_SCROLLLOCK);

    if(Keyboard.ScrollLock == 1)
      Keyboard.ScrollLock  = 0;
    else
      Keyboard.ScrollLock  = 1;

    return 0;
  }

  if(keycode >= 0x80)
    return 0;

  if(Keyboard.shift > 0)
    c = kbmap2[keycode];
  else
    c = kbmap1[keycode];

  if(((Keyboard.CapsLock == 1 && Keyboard.shift == 1) ||
     (Keyboard.CapsLock == 0 && Keyboard.shift == 0)) &&
     ((c >= 0x41 && c <= 0x5a) || (c >= 0x61 && c <= 0x7a)))
      c += 0x20;

  return c;
}



/*
*  Function: keyboardLED
*
*    キーボードのLEDを制御する
*
*  Parameters:
*
*    target  - 点灯または消灯させるLEDのキーコード
*
*  Returns:
*
*    なし
*
*/

void keyboardLED (unsigned char target){

 u8 data;


  if(target == 0)
    data = 0;

  if(target == KEY_CAPSLOCK)
    data = (((Keyboard.CapsLock)<<2 == 0 ) ? 1<<2:0)+((Keyboard.NumLock)<<1) +(Keyboard.ScrollLock);
  if(target == KEY_NUMLOCK)
    data = (((Keyboard.NumLock)<<1 == 0 ) ?  1<<1:0)+((Keyboard.CapsLock)<<2)+(Keyboard.ScrollLock);
  if(target == KEY_SCROLLLOCK)
    data = (((Keyboard.ScrollLock) == 0 ) ?     1:0)+((Keyboard.CapsLock)<<2)+((Keyboard.NumLock)<<1);


  do{

    waitKBC(1);
    out8(IOPORT_KBC1, 0xed);
    waitKBC(0);

  }while(in8(0x60) == 0xfe);



  do{

    out8(IOPORT_KBC1, data);
    waitKBC(0);

  }while(in8(0x60) == 0xfe);



  return;
}




/*
*  Function: waitKBC
*
*    キーボードコントローラが受信可能な状態になるまで待つ
*
*  Parameters:
*
*    bit  - 対象となるビット
*
*  Returns:
*
*    なし
*
*/

void waitKBC (int bit){

 char stateReg;

  do{

    stateReg = in8(IOPORT_KBC2);

  }while((stateReg&0x02)>>bit != 0);

  return;
}




int getchar(void){

 int ch2;

  while(keybuf == 0) waitInterrupt();

  ch2 = (int) keybuf;
  keybuf = 0;

  return ch2;
}



void puts (const char *str){

 int i;

  for(i=0; str[i] != '\0'; i++)
    putchar(str[i]);

  putchar('\n');

  return;
}




int strcmp (const char *str1, const char *str2){

 int i;

  for(i=0; (str1[i] == str2[i]) &&
           (str1[i] != '\0')    &&
           (str2[i] != '\0')    ;
      i++);

  return ((str1[i] == '\0') && (str2[i] == '\0'))? 0:1;

}


